int
prioq_put(struct prioq *q, struct prioq_head *h)
{
	int idx, parent_idx;

	if (q->used == q->size)
		prioq_resize(q, q->size + 32);

	idx = q->used++;

	while (idx > 0) {

		parent_idx = PARENT(idx);

		if (q->compare(h, q->heap[parent_idx]) < 0)
			break;

		q->heap[parent_idx]->index = idx;
		q->heap[idx] = q->heap[parent_idx];
		idx = parent_idx;
	}

	q->heap[idx] = h;
	h->index = idx;

	return 0;
}

static __inline__ void
prioq_heapify(struct prioq *q, int idx)
{
	int idx_child_max;
	struct prioq_head *tmp;
	int idx_right, idx_left;

	while ((idx_left = LEFT_CHILD(idx)) < q->used) {

		idx_right = RIGHT_CHILD(idx);

		if ((idx_right < q->used) &&
		    (q->compare(q->heap[idx_right], q->heap[idx_left]) > 0))
			idx_child_max = idx_right;
		else
			idx_child_max = idx_left;

		tmp = q->heap[idx];
		q->heap[idx] = q->heap[idx_child_max];
		q->heap[idx]->index = idx;
		q->heap[idx_child_max] = tmp;
		q->heap[idx_child_max]->index = idx_child_max;
		idx = idx_child_max;
	}
}

struct prioq_head *
prioq_get(struct prioq *q)
{
	struct prioq_head *h;

	if (q->used == 0)
		return NULL;

	h = q->heap[0];

	q->used--;
	q->heap[0] = q->heap[q->used];
	q->heap[0]->index = 0;

	prioq_heapify(q, 0);

	return h;
}

void prioq_remove(struct prioq *q, struct prioq_head *h)
{
	int idx, parent_idx;

	idx = h->index;
	h->index = -1;

	q->used--;

	if (q->used > 0 && idx != q->used) {

		while (idx > 0) {
			parent_idx = PARENT(idx);

			if (q->compare(q->heap[q->used], q->heap[parent_idx]) < 0)
				break;

			q->heap[idx] = q->heap[parent_idx];
			q->heap[idx]->index = idx;
			idx = parent_idx;
		}

		q->heap[idx] = q->heap[q->used];
		q->heap[idx]->index = idx;

		prioq_heapify(q, idx);
	}
}
